import numpy as np
import pandas as pd
import copy as copy
from scipy.linalg import inv
from scipy.sparse import diags
from scipy.sparse import csc_matrix
import time


def group_prox(x, t, groups, weights):
    x0 = x.copy()
    for g, w in zip(groups, weights):
        xn = np.linalg.norm(x0[g])
        r = t * w
        if xn > r:
            x0[g] -= x0[g] * r / xn
        else:
            x0[g] = 0
    return x0


def generate_groups(p, q):
    num_rows = int(p)
    c = np.empty([num_rows, q], dtype=int)
    for i, _ in enumerate(range(num_rows)):
        c[i] = np.arange(i*q, (i+1)*q)
    return c


def BF(df_log, lambda_val, rho=10., eps=1e-12, max_iterations=10000):
    df_log0 = df_log.copy()
    Y = df_log.copy().to_numpy()
    p = Y.shape[1]
    n = Y.shape[0]
    groups = generate_groups(p-1, n)
    weights = np.ones(groups.shape[0])*np.sqrt(n)
    N0 = diags([1, -1], [0, 1], shape=(p-1, p))
    D = N0.T
    DDT = D@D.T

    def Loss(Y, B):
        a = Y - B
        return(np.sum(a*a))

    a = inv(np.identity(p) + rho/2*DDT)
    a = np.round(a, 6)
    a = csc_matrix(a)
    D = csc_matrix(D)
    DDT = csc_matrix(DDT)

    def B_funct(Y, G, U, rho):
        return((Y + rho/2*(G-U/rho)@D.T)@a)

    def G_funct(B, U, lamb, rho):
        bd = (B@D).T.flatten()
        u = U.T.flatten()
        bd += u/rho
        temp = group_prox(bd, lamb/rho, groups, weights).reshape((p-1, n))
        return(temp.T)

    def U_funct(U, B, G, rho):
        U += rho*(B@D - G)
        return(U)

    start = time.time()
    iterations = max_iterations
    Gk_1 = np.zeros((n, p-1))
    Uk_1 = np.zeros((n, p-1))
    Bk = np.zeros((n, p))
    j = 0
    lamb = lambda_val
    for i in range(0, iterations):
        Bk_old = Bk
        Bk = B_funct(Y, Gk_1, Uk_1, rho)
        Gk = G_funct(Bk, Uk_1, lamb, rho)
        Uk = U_funct(Uk_1, Bk, Gk, rho)
        Gk_1 = Gk
        Uk_1 = Uk
        err = np.linalg.norm(Bk - Bk_old, 'fro')/np.linalg.norm(Bk, 'fro')
        if i % 1000 == 0:
            print('lambda = %1.2e, i= %i, err = %1.2e' % (lamb, i, err))
        if i > 10 and err < eps:
            break
    print('lambda = %1.2e, i= %i, err = %1.2e' % (lamb, i, err))

    df_log0.iloc[:,:] = Bk
    print('running time:', (time.time() - start))

    return(df_log0)


def reduce_merge_lists(ids):
    ids = np.r_[ids, [[-1,-1]]]
    ids_new = np.zeros((0,2))
    start, stop = ids[0,:]
    nrow = ids.shape[0]
    for i in range(nrow-1):
        if not ids[i,1]==ids[i+1,0]:
            ids_new = np.r_[ids_new, [[start,stop]]]
            start, stop = ids[i+1,:]
        else:
            stop = ids[i+1,1]
    return(ids_new)


def BF_aggregate_bins(df_BF, df_original, round_digits=5, min_bin_length=1):
    Y_BF = df_BF.to_numpy()
    Y_original = df_original.to_numpy()
    Y_BF = np.round(Y_BF, decimals=round_digits)
    p = Y_BF.shape[1]
    n = Y_BF.shape[0]
    current = Y_BF[:, 0]
    plateau_id = np.zeros((0, 2))
    non_plateau_id = np.zeros((0, 2))

    for i in range(0, p-1):
        new = Y_BF[:, i+1]
        if np.equal(current, new).all():
            plateau_id = np.r_[plateau_id, [[i, i+1]]]
        else:
            non_plateau_id = np.r_[non_plateau_id, [[i, i+1]]]
        current = Y_BF[:, i+1]

    plat_id = reduce_merge_lists(plateau_id)
    non_plat_id = reduce_merge_lists(non_plateau_id)

    df_plat_id = pd.DataFrame(data=plat_id,
                              columns=["start", "stop"]).astype(int)
    df_plat_id = df_plat_id.astype(int)
    df_plat_id['plateau'] = pd.Series("yes", index=df_plat_id.index)
    df_non_plat_id = pd.DataFrame(data=non_plat_id, columns=["start", "stop"])
    df_non_plat_id = df_non_plat_id.astype(int)
    df_non_plat_id['plateau'] = pd.Series("no", index=df_non_plat_id.index)
    df_non_plat_id['start'] = df_non_plat_id['start'] + 1
    df_non_plat_id['stop'] = df_non_plat_id['stop'] - 1
    df_id = pd.concat([df_plat_id, df_non_plat_id], ignore_index=True)
    df_id = df_id[pd.Series(df_id['stop']-df_id['start']+1
                            >=min_bin_length).values]
    df_id = df_id.sort_values(by = 'start',
                              ascending=True).reset_index(drop=True)
    df_id['start bin'] = df_BF.columns[df_id['start']]
    df_id['stop bin'] = df_BF.columns[df_id['stop']]
    df_id['start to stop bin'] = df_BF.columns[df_id['start']] + "-" + df_BF.columns[df_id['stop']]

    Y_new_BF = np.zeros((n, df_id.shape[0]))
    Y_new_Orig = np.zeros((n, df_id.shape[0]))

    for i in range(df_id.shape[0]):
        #print(Y_BF[0:1,
        #            df_id['start'][i]:(1+df_id['stop'][i])])
        Y_new_BF[:,i] = np.mean(Y_BF[:,
                    df_id['start'][i]:(1+df_id['stop'][i])], axis=1)
        Y_new_Orig[:,i] = np.sum(Y_original[:,
                    df_id['start'][i]:(1+df_id['stop'][i])], axis=1)

    Y_new_BF_df = pd.DataFrame(data=Y_new_BF,
                              columns = df_id['start to stop bin'],
                              index = df_BF.index)

    Y_new_Orig_df = pd.DataFrame(data=Y_new_Orig,
                              columns = df_id['start to stop bin'],
                              index = df_BF.index)
    return(Y_new_BF_df, Y_new_Orig_df, df_id)


def apply_BF_model(df_original, df_id):
    Y_original = df_original.to_numpy()
    n = Y_original.shape[0]
    Y_new_Orig = np.zeros((n, df_id.shape[0]))
    for i in range(df_id.shape[0]):
        Y_new_Orig[:,i] = np.sum(Y_original[:,
                    int(df_id['start'][i]):int((1+df_id['stop'][i]))], axis=1)

    Y_new_Orig_df = pd.DataFrame(data=Y_new_Orig,
                              columns = df_id['start to stop bin'],
                              index = df_original.index)
    return(Y_new_Orig_df)


def BF_wrapper(df_original, lambda_val, rho=10., round_digits=5, eps=1e-12, max_iterations=10000):
    df_log = np.log2(df_original)

    df_BF = BF(df_log, lambda_val, rho=rho, eps=eps, max_iterations=max_iterations)

    df_log_binned, df_original_binned, BF_summary = BF_aggregate_bins(df_BF, df_original,
                                                      round_digits=round_digits)

    return df_original_binned, BF_summary
